#!/usr/bin/env python
# encoding: utf-8
"""machinery for compiling new versions of appimages."""

import datetime
import json
import loaih.solvers
from packaging.version import parse as parse_version

class Definitions(): # pylint: disable=too-few-public-methods
    """Definitions for the module."""

    DOWNLOADPAGE = "https://www.libreoffice.org/download/download/"
    ARCHIVE = "https://downloadarchive.documentfoundation.org/libreoffice/old/"
    RELEASE = "https://download.documentfoundation.org/libreoffice/stable/"
    DAILY = "https://dev-builds.libreoffice.org/daily/master/"
    PRERELEASE = "https://dev-builds.libreoffice.org/pre-releases/deb/x86_64/"

    SELECTORS = {
        'still': {
            'URL': DOWNLOADPAGE,
            'xpath': '(//span[@class="dl_version_number"])[last()]/text()'
        },
        'fresh': {
            'URL': DOWNLOADPAGE,
            'xpath': '(//span[@class="dl_version_number"])[1]/text()'
        },
        'prerelease': {
            'URL': DOWNLOADPAGE,
            'xpath': '//p[@class="lead_libre"][last()]/following-sibling::ul[last()]/li/a/text()'
        },
        'daily': {
            'URL': DAILY,
            'xpath': '//td/a'
        }
    }


class Version():
    """Represent the skeleton of each queried version."""

    def __init__(self):
        self.query = ''
        self.version = ''
        self.urls = {
            'x86': '-',
            'x86_64': '-'
        }

    def to_dict(self):
        """Returns a dictionary of versions."""
        return {
            'query': self.query,
            'version': self.version,
            'basedirurl': self.urls
        }

    def to_json(self):
        """Returns a json representation of the version."""
        return json.dumps(self.to_dict())

    def __str__(self):
        return f"""query: {self.query}
version: {self.version}
x86: {self.urls['x86']}
x86_64: {self.urls['x86_64']}"""

class QueryError(Exception): pass

class Query():
    """Represents each query and helps determining the aspects of a version."""

    def __init__(self, query: str, default_to_current = False):
        self.text = query
        self.type = ''
        self.default_to_current = default_to_current

        # Let's first determine which type of query we are doing.
        if self.text in { 'current', 'yesterday', 'daily' }:
            self.type = 'daily'
        elif self.text in { 'still', 'fresh', 'prerelease' }:
            self.type = 'named'
        elif '.' in self.text:
            self.type = 'exact_version'
        else:
            try:
                int(query)
            except ValueError:
                raise QueryError("The queried version does not exist.")

        # Since the query is number only, let's let it leave with the
        # number, but set the type to 'daily'
        self.type = 'daily'

        # call Base.parse_query to popolate the results
        self.results = loaih.solvers.Solver.parse_query(self)


class Base():
    """Contains methods that might be useful outside class."""
    # Class for static methods which might be useful even outside the build
    # scripts.

    @staticmethod
    def dailyurl(date = datetime.datetime.today(), default_current=False):
        """Returns the URL for the latest valid daily build."""
        # As per other parts of the build, we need to maintain an URL also for
        # x86 versions that it isn't really provided.
        # As such, the return value must be a dictionary

        # Fixing daily selector
        # As seen, the number of the tinderbox building the daily version can
        # change. We try to fulfill the void by adding a step.
        tinderboxpage = etree.HTML(#pylint: disable=c-extension-no-member
            urllib.request.urlopen(Definitions.DAILY).read() #pylint: disable=consider-using-with
        )
        xpath = "//td/a[starts-with(text(), 'Linux-rpm_deb-x86') and contains(text(), 'TDF/')]/text()" #pylint: disable=line-too-long
        tburl = str(tinderboxpage.xpath(xpath)[0])
        daily_selector = f"{Definitions.DAILY}{tburl}"

        # Get the anchor for today's builds
        raw_page = etree.HTML(urllib.request.urlopen(daily_selector).read())# pylint: disable=c-extension-no-member,consider-using-with


        results = raw_page.xpath(
            f"""//td/a[contains(text(), "{date.strftime('%Y-%m-%d')}")]/text()"""
        )
        if len(results) > 0:
            # On the contrary, more than a version is found. let's order the
            # list and get the latest item
            return { 'x86': '-', 'x86_64': f"{daily_selector}{sorted(results)[-1]}" }
        # No results found, no version found.
        # But if default_current is true, redo the search with 'current'
        if not default_current:
            return { 'x86': '-', 'x86_64': '-' }

        # default_current is true - redo all the queries for current
        current_link = raw_page.xpath("//td/a[contains(text(), 'current']/text()")[0]
        return {'x86': '-', 'x86_64': f"{daily_selector}{current_link}" }


    @staticmethod
    def dailyver(date = datetime.datetime.today(), force_current = False):
        """Returns versions present on the latest daily build."""
        url = Base.dailyurl(date, force_current)['x86_64']
        # If no daily releases has been provided yet, return empty
        if url == '-':
            return []

        # Rerun the page parsing, this time to find out the versions built
        fullpage = urllib.request.urlopen(url).read() #pylint: disable=consider-using-with
        archive_path = "//td/a[contains(text(), '_deb.tar.gz')]/text()"
        tarball = etree.HTML(fullpage).xpath(archive_path)#pylint: disable=c-extension-no-member
        # This should have returned the main package for a version, but can
        # have returned multiple ones, so let's treat it as a list
        tarball_versions = [ x.split('_')[1] for x in tarball ]
        if len(tarball_versions) == 1:
            return tarball_versions[0]

    @staticmethod
    def namedver(query):
        """Gets the version for a specific named version."""

        if 'current' in query:
            # Should return the daily version with the current link
            return Base.dailyver(datetime.datetime.today(), True)

        if 'yesterday' in query:
            return Base.dailyver(datetime.datetime.now() + datetime.timedelta(days=-1))

        if 'daily' in query:
            return Base.dailyver()

        # If the name is convertible to integer, it means it is written as
        # <year><month><date> and we ask for a daily build.
        try:
            int(query)
        except ValueError:
            # All other options - fresh, prerelease, still
            return etree.HTML(#pylint: disable=c-extension-no-member
                urllib.request.urlopen(
                    Definitions.SELECTORS[query]['URL']
                ).read()
            ).xpath(Definitions.SELECTORS[query]['xpath'])

        # Lets restart with integer version
        return Base.dailyver(datetime.datetime.strptime(query, "%Y%m%d"))

    @staticmethod
    def fullversion(version):
        """Get latest full version from Archive based on partial version."""
        versionlist = etree.HTML(urllib.request.urlopen(Definitions.ARCHIVE).read()).xpath(f"//td/a[starts-with(text(), '{version}')]/text()")
        if versionlist:
            cleanlist = sorted([ x.strip('/') for x in versionlist ])

            # Sorting, then returning the last version
            return cleanlist[-1]

        return None

    @staticmethod
    def urlfromqueryandver(query, version):
        """Returns the fetching URL based on the queried version and the numeric version of it."""
        # This has the purpose to simplify and explain how the releases are
        # layed out.

        # If the query tells about daily or 'yesterday' (for testing purposes),
        # we might ignore versions and return the value coming from dailyurl:
        if query == 'daily':
            return Base.dailyurl()
        if query == 'yesterday':
            date = datetime.datetime.today() + datetime.timedelta(days=-1)
            return Base.dailyurl(date)

        # All other versions will be taken from Archive, as such we need a full
        # version.

        # If the version has only 2 points in it (or splits into three parts by '.'), that's not a full version and we will call the getlatestver() function
        fullversion = str(version)
        if len(fullversion.split('.')) <= 3:
            fullversion = str(Base.fullversion(version))

        # So the final URL is the Archive one, plus the full versions, plus a
        # final '/deb/' - and an arch subfolder
        baseurl = Definitions.ARCHIVE + fullversion + '/deb/'
        retval = {}

        # x86 binaries are not anymore offered after 6.3.0.
        if parse_version(fullversion) < parse_version('6.3.0'):
            retval['x86'] = baseurl + 'x86/'
        else:
            retval['x86'] = '-'

        retval['x86_64'] = baseurl + 'x86_64/'

        return retval

    #@staticmethod
    #def collectedbuilds(query, default_current=False):
    #    """Creates a list of Builds based on each queried version found."""
    #    retval = []
    #    if '.' in query:
    #        # Called with a numeric query. Pass it to RemoteBuild
    #        retval.append(RemoteBuild(query))
    #    else:
    #        try:
    #            int(query)
    #        except ValueError:
    #            # Named query
    #            named_version = Base.namedver(query)
    #            # named_version should be just one version anyway
    #            if isinstance(named_version, list) and len(named_version) > 1:
    #                for ver in named_version:
    #                    if 'daily' in query:

    #            if 'daily' in query and specs == {}:
    #                # No daily build has been found.
    #                # Rerun the query but with 'current'
    #                named_version = Base.namedver('current')

    #            if specs != {}:
    #                if isinstance(specs, list) and len(specs) > 1:
    #                    for spec in specs:
    #                        remotebuild = RemoteBuild(query)
    #                        if 'daily' in query:
    #                            remotebuild.daily(spec)
    #                        else:
    #                            remotebuild.from_version(spec['version'])

    #                        retval.append(remotebuild)

    #                else:
    #                    # Possibly single instance of dict
    #                    remotebuild = RemoteBuild(query)
    #                    remotebuild.from_version(specs['version'])
    #                    retval.append(remotebuild)
    #            else:
    #                return retval

    #        # If the query is convertible in integer, we are still asking for a
    #        # daily build in a specific day
    #        specs = Base.dailyver(datetime.datetime.strptime(query, '%Y%m%d'))
    #        if isinstance(specs, list) and len(specs) > 1:
    #            for spec in specs:
    #                remotebuild = RemoteBuild('daily')
    #                remotebuild.daily(spec)
    #                retval.append(remotebuild)
    #        else:
    #            remotebuild = RemoteBuild('daily')
    #            remotebuild.daily(specs)
    #            retval.append(remotebuild)

    #    return sorted(retval, key=lambda x: x.version)


class RemoteBuild():
    """Builds a version with checking remotely if it was not already built."""

    def __init__(self, query, version = None):
        """Should simplify the single builded version."""
        self.query = query
        self.version = version or ''
        self.basedirurl = { 'x86': '-', 'x86_64': '-' }

    def from_version(self, version):
        if version and isinstance(version, str):
            self.version = version

        if not '.' in self.query:
            # Named version.
            # Let's check if a specific version was requested.
            if self.version == '':
                # In case it was not requested, we will carry on the generic
                # namedver() query.
                # If the results are more than one, we'll take the latest
                # (since we are requested to provide a single build).
                a_version = Base.namedver(self.query)

                if isinstance(a_version, list):
                    # if the number of versions is zero, return and exit
                    if not a_version:
                        self.version = None

                    if len(a_version) == 1:
                        # version is a single one.
                        self.version = a_version[0]
                    else:
                        # In this case, we will select the latest release.
                        self.version = sorted(a_version)[-1]

            # If the version has already a version, as requested by user,
            # continue using that version
        else:
            # In case of numbered queries, put it as initial version
            self.version = self.query

        if len(str(self.version).split('.')) < 4:
            # If not 4 dotted, let's search for the 4 dotted version
            self.version = Base.fullversion(self.version)

        self.basedirurl = Base.urlfromqueryandver(self.query, self.version)

    def daily(self, dailyspecs):
        """Builds a remote version starting from dailyver dictionary."""
        self.version = dailyspecs['version']
        self.basedirurl = dailyspecs['basedirurl']


