#!/usr/bin/env python
# encoding: utf-8
"""Solvers for the queries."""

import datetime
import requests
from lxml import html

class Solver():
    """Helps solving the queries to collections of versions."""

    @staticmethod
    def parse_query(query: Query):
        """Returns a list of versions for the query."""

        retval = []
        if query.type == 'daily':
            solver = DailySolver(query.text)
            retval.append(solver.to_version())
        elif query.type == 'named':
            solver = NamedSolver(query.text)
        else:
            solver = NumberedSolver(query.text)

        return retval


class DailySolver():
    def __init__(self, query, default_to_current = False):
        self.query = query
        self.default_to_current = default_to_current
        self.version = ''
        self.url = ''
        self.__get_url__()
        self.__get_version__()

    def __get_url__(self):
        """Get daily urls based on query."""
        # The base URL for daily releases is already determined. Let's define
        # the definitive one.

        if self.query != 'current':
            baseurl = requests.get(Definitions.DAILY)
            pageobj = html.fromstring(baseurl.content)
            xpath_string = "//td/a[starts-with(text(), 'Linux-rpm_deb-x86') and contains(text(), 'TDF/')]/text()"
            tinderbox_segment = str(pageobj.xpath(xpath_string)[-1])
            baseurl= f"{Definitions.DAILY}{tinderbox_segment}"

            # Reiterate now to search for the dated version
            base_page = requests.get(baseurl)
            baseobj = html.fromstring(base_page.content)
            daily_set = baseobj.xpath("//td/a/text()")
            
            searchdate = datetime.datetime.today()
            if self.query == 'yesterday':
                searchdate = searchdate + datetime.timedelta(days=-1)
            else:
                searchdate = datetime.datetime.strptime(self.query, '%Y%m%d')

            search_results = [ x for x in daily_set if searchdate.strftime('%Y-%m-%d') in x ][-1]

            if len(search_results) < 1:
                # Searched date do not exist. if default_to_current is set,
                # let's re-run the solver with 'current' query.
                if self.default_to_current:
                    current = DailySolver('current')
                    current.query = self.query
                    self.url = current.url

            else:
                # We'll presume there will be just one result anyways.
                self.url = f"{baseurl}{search_results[-1]}"

        else:
            # Current.
            current_page = requests.get(f"{Definitions.DAILY}current.html")
            current_obj = html.fromstring(current_page.content)
            xpath_string = "//td/a[contains(@href, 'Linux-rpm_deb-x86') and contains(@href, 'TDF/') and contains(@href, 'deb.tar')]/@href"
            current_link = str(current_obj.xpath(xpath_string)[-1])

            split_link = str(current_link).split('/')
            return '/'.join(split_link[1:-1]) + '/'

    def __get_version__(self):
        page = requests.get(self.baseurl)
        obj = html.fromstring(page.content)
        xpath_string = "//td/a[contains(text(), '_deb.tar.gz')]/text()"
        link = str(obj.xpath(xpath_string)[-1])
        self.version = link.split('/')[-1].split('_')[1]

    def to_version(self):
        version = Version()
        version.query = self.query
        version.version = self.version
        version.urls['x86_64'] = self.url
        return version
